import pathlib
from typing import List, Optional, Set

from pyspark_data_mocker import config, utils
from pyspark_data_mocker.spark_session import SparkTestSession
from pyspark_data_mocker.utils import PathLike


class DataLakeBuilder:
    def __init__(self, spark_test: SparkTestSession, schema_configuration: Optional[PathLike] = None):
        self.dbs: Set[str] = set()
        self.tables: List[dict] = list()
        self.spark_test = spark_test
        self.spark = spark_test.session

        self.schema: Optional[dict] = None
        if schema_configuration:
            schema_configuration = utils.to_path(schema_configuration)
            self.schema = config.get_schema_configuration_from_dir(schema_configuration)

    def with_db(self, name: str) -> "DataLakeBuilder":
        """
        Register a database to be created when executing the plan. Ignore database already registered.

        :param name:    STRING  Name of the database to be created
        :return:        An instance of the DataLakeMocker modified, to be able to chain methods
        """
        if name in self.dbs:
            return self
        self.dbs.add(name)
        return self

    def with_table(self, table_name: str, fmt: str, path: PathLike, db_name: str = "default") -> "DataLakeBuilder":
        """
        Register a new table in the datalake. The creation of this table will be later when executing the whole
        plan. If <path> is provided, then the table will be loaded reading that file. It is mandatory that if you
        set a file <path> then, a format should be provided.

        TODO:   in further features, a possiblity to pass as argument a TableBuilder or similar will be possible,
                to be able to define the table using a DSL
        :param table_name:  STRING          The name of the table that will be created
        :param fmt:         STRING          'csv', 'parquet', 'text' or 'json'
        :param path:        STRING or Path  location of the file
        :param db_name:     STRING          name of the database where the table will leave. Default value: 'default'
        :return:            An instance of the DataLakeMocker modified, to be able to chain methods
        """
        # TODO: make a dataclass of the table
        self.tables.append({"path": str(path), "format": fmt, "table_name": table_name, "db_name": db_name})
        return self

    def run(self) -> "DataLakeBuilder":
        """
        Executes the plan by creating each database and table configured

        :return:    The reference of the builder
        """
        for db in self.dbs:
            self.spark.sql(f"CREATE DATABASE IF NOT EXISTS {db}")

        for table in self.tables:
            # TODO: configure schema infering
            opts = dict(header=True, inferSchema=False)
            table_full_name = f"{table['db_name']}.{table['table_name']}"
            reader = self.spark.read
            if self.schema and table_full_name in self.schema:
                reader = reader.schema(utils.dict_to_ddl_string(self.schema[table_full_name]))
            df = reader.format(table["format"]).options(**opts).load(table["path"])
            df.write.mode("overwrite").saveAsTable(table_full_name)

        return self

    def cleanup(self):
        for table in self.tables:
            self.spark.sql(f"TRUNCATE TABLE {table['db_name']}.{table['table_name']}")
            self.spark.sql(f"DROP TABLE {table['db_name']}.{table['table_name']}")
        for db in self.dbs:
            self.spark.sql(f"DROP DATABASE IF EXISTS {db}")

    @staticmethod
    def load_from_dir(datalake_dir: PathLike, app_config_path: Optional[PathLike] = None) -> "DataLakeBuilder":
        """
        Navigates over the <datalake_dir> to create the datalake automatically. The file structure needs to be like
        this:
            * The root folder should contain table_like data (csv files, parquet files, json files, all files that
            can be read and be interpreted as a dataframe) or folders.
                * if the <datalake_dir> contains at the root level table-like data, those will be considered as tables
                in the 'default' database.
                * All directories inside the <datalake_dir> folder are considered as other databases, and inside
                the directory it should only contain table_like data

        :param datalake_dir: Directory that contains the datalake definition (table-like files to load as tables and/or
                             folders that will be considered as databases)
        :param app_config_path:   Optional argument with a path of a yaml file to configure the spark session to use
        """
        datalake_dir = utils.to_path(datalake_dir)
        if app_config_path:
            app_config = config.get_config_from_dir(app_config_path)
        else:
            app_config = config.default_config()
        if not datalake_dir.exists():
            raise ValueError(f"The path provided '{datalake_dir}' does not exists")

        if not datalake_dir.is_dir():
            raise ValueError(f"The path '{datalake_dir}' is not a directory with a delta lake data")

        spark_test = SparkTestSession(app_config)
        schema_config_path = pathlib.Path(datalake_dir, app_config.schema_config_file_name)
        schema_config = schema_config_path if schema_config_path.exists() else None
        builder = DataLakeBuilder(spark_test, schema_configuration=schema_config)

        for d in datalake_dir.iterdir():
            if d.name == app_config.schema_config_file_name:
                continue

            if d.is_file():
                table_name, extension = d.name.split(".")
                builder = builder.with_table(table_name=table_name, fmt=extension, path=d.resolve())
            else:
                db_name = d.name
                builder = builder.with_db(db_name)
                for table in d.iterdir():
                    table_name, extension = table.name.split(".")
                    builder = builder.with_table(
                        table_name=table_name, fmt=extension, path=table.resolve(), db_name=db_name
                    )

        builder.run()
        return builder



from typing import Callable


def range_between(min_value: int, max_value: int) -> Callable:
    """
    Returns a callback that checks if the given configuration number is between the configured range. Raises
    an error that will be caught later in schema module.

    :param min_value:
    :param max_value:
    :return:
    """
    if min_value > max_value:
        raise Exception(f"Bad schema configuration: range is not valid ({min_value=}, {max_value=})")

    def validate(n: int):
        if not min_value <= n <= max_value:
            raise Exception(f"Number configured is not between the range [{min_value}, {max_value}]")
        return True

    return validate


def validate_version(valid_versions: set) -> Callable:
    """
    Given a set of valid versions, return a callback that will be used later for schema validation that raises
    and error with a readable description for the user if the version configured is not valid.
    :param valid_versions:
    :return:
    """

    def validate(version: str):
        if version not in valid_versions:
            raise Exception(f"Version '{version}' it not in the list of supported versions ({sorted(valid_versions)})")
        return True

    return validate



import logging
from tempfile import TemporaryDirectory
from typing import Union

from pyspark.sql import SparkSession

from pyspark_data_mocker.config import AppConfig

Dir = Union[str, TemporaryDirectory]


class SparkTestSession:
    def __init__(self, config: AppConfig):
        self.log = logging.getLogger()
        self.config = config
        self._session = self.__create_session(config)

    def __create_session(self, config: AppConfig) -> SparkSession:
        """
        Creates a spark session with configuration that improves the local setup and
        execution. This is intended because the datalake that will be mocked for tests will have usually low
        amount of data. If we use the default spark configuration its probably that the tests takes too long
        to run due to unnecessary optimizations for big data.

        Special Thanks to Sergey Ivanychev for his research on "Faster PySpark Unit Test"
        [blog](https://medium.com/constructor-engineering/faster-pyspark-unit-tests-1cb7dfa6bdf6)

        :param config: Application configuration
        :return: a new spark session configured to be performant for tests
        """
        self.log.info(f"Creating session of name '{config.app_name}' with {config.number_of_cores} cores")
        builder = (
            SparkSession.builder.master(f"local[{config.number_of_cores}]")
            .appName(config.app_name)
            .config("spark.sql.warehouse.dir", config.spark_warehouse_dir_path)
        )

        self.log.info(f"Spark warehouse dir: {config.spark_warehouse_dir_path}")
        # The default value of this config is 200, that is too much for local tests that will not handle
        # so much data.
        builder = builder.config("spark.sql.shuffle.partitions", 1)

        # accumulate all JVM options and at the end set the JVM configuration
        # -XX:+CMSClassUnloadingEnabled: allows JVM garbage collector to remove unused classes that spark generates
        # -XX:+UseCompressedOops: tells JVM to use 32-bit addresses instead of 64 (if you don't use more that 32GB of
        #                         ram there would not be any problem
        jvm_options = ["-XX:+CMSClassUnloadingEnabled", "-XX:+UseCompressedOops"]

        # Enable delta optimizations if configured
        if config.delta_configuration:
            dconfig = config.delta_configuration
            self.log.info(f"Enabling delta configuration using delta version '{dconfig.delta_version}'")
            delta_package = f"io.delta:delta-core_{dconfig.scala_version}:{dconfig.delta_version}"
            builder = (
                builder.config("spark.jars.packages", delta_package)
                .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
                .config("spark.databricks.delta.snapshotPartitions", dconfig.snapshot_partitions)
                .config("spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog")
            )
            jvm_options.append(f"-Ddelta.log.cacheSize={dconfig.log_cache_size}")

        builder = (
            # disable console progress bar, TODO: maybe add configure this using the config file
            builder.config("spark.ui.showConsoleProgress", "false")
            # Disable spark ui to reduce RAM and lower the local HTTP request to the server
            .config("spark.ui.enabled", "false")
            .config("spark.ui.dagGraph.retainedRootRDDs", "1")
            .config("spark.ui.retainedJobs", "1")
            .config("spark.ui.retainedStages", "1")
            .config("spark.ui.retainedTasks", "1")
            .config("spark.sql.ui.retainedExecutions", "1")
            .config("spark.worker.ui.retainedExecutors", "1")
            .config("spark.worker.ui.retainedDrivers", "1")
        )

        builder = builder.config("spark.driver.extraJavaOptions", " ".join(jvm_options))

        # Set a low memory ram to control the memory usage. Consider that this will affect the master, and
        # all cores used (the RAM consumed will be this value times the amount of core used)
        # TODO: parametrize this
        builder = builder.config("spark.driver.memory", "1g")
        if config.enable_hive:
            self.log.info("Enabling Hive support")
            builder = builder.enableHiveSupport()

        return builder.getOrCreate()

    @property
    def session(self) -> SparkSession:
        return self._session



import pathlib
import re

import yaml
from dacite import from_dict

from pyspark_data_mocker.config.app_config import AppConfig
from pyspark_data_mocker.config.schema import config_schema
from pyspark_data_mocker.utils import PathLike


def get_config_from_dir(filepath: PathLike) -> AppConfig:
    """
    Reads the configuration of the yaml file located in <filepath>, validates it and return an
    AppConfig dataclass instance of the app configuration.

    :param filepath:
    :return:
    """
    with open(filepath, "r") as file:
        config = yaml.safe_load(file)

    config = config_schema.validate(config)
    return from_dict(data_class=AppConfig, data=config)


def get_schema_configuration_from_dir(filepath: PathLike) -> dict:
    """
    Given the path to a schema_config.yaml file, loads the file, and validates that the file defines table schemas

    :param filepath:
    :return:
    """
    import schema

    with open(filepath, "r") as file:
        config = yaml.safe_load(file)

    # The schema of the file is dynamic, the only restriction is that the config file must contain a dict of dicts,
    # where each key must contain a string value (later we can add more restrictions like the latter str must be a valid
    # DDL-formatted string

    # TODO:   due to an issue with schema (https://github.com/keleshev/schema/issues/255) it is not possible to make a
    #        user friendly error message here. I will add the check with a proper error msg for the keys outside, but
    #        when that bug is fixed, it should be added here using a Regex
    #        schema.Regex(
    #            r"^[a-z]+\.[a-z]+$",
    #            flags=re.I,
    #            error="Table name must contain the form <database>.<table> name convention",
    #        )
    schema_validator = schema.Schema(
        {table_name: {name: str for name in table_config.keys()} for table_name, table_config in config.items()}
    )
    config_validated = schema_validator.validate(config)

    for table_name in config_validated.keys():
        if not re.search(r"^[a-z]+\.[a-z]+$", table_name, re.IGNORECASE):
            raise ValueError(
                f"Invalid parameter: Table '{table_name}' name must follow the form <database>.<table> name convention"
            )
    return config_validated


def default_config() -> AppConfig:
    """
    :return: Returns a default AppConfig class with basic configuration
    """
    default_config_path = pathlib.Path(pathlib.Path(__file__).parent, "default_config.yaml")
    return get_config_from_dir(default_config_path)



from pathlib import Path
from typing import Union

PathLike = Union[str, Path]


def to_path(path: PathLike) -> Path:
    """
    Transforms a PathLike class (that can be a pathlib.Path or a string) into a pathlib.Path

    :param path: PathLike   the path to transform
    :return:
    """
    if isinstance(path, Path):
        return path

    return Path(path)


def dict_to_ddl_string(table_definition: dict) -> str:
    """
    Given a table definition, make a string with the ddl for pyspark, to be able to infer the schema of the dataframe

    Example:
        Considering this table definition:
            {
                "id": "int",
                "start_date": "date",
                "metadata": "map<string, string>",
                "name": "string"
            }
        this function will return
        "id string, start_date date, metadata map<string, string>, name string"
    :param table_definition:
    :return:
    """
    rows = [f"{col_name} {col_type}" for col_name, col_type in table_definition.items()]
    return ", ".join(rows)
